@@ -23,7 +23,6 @@
 #include "service-type.h"
 #include "shell.h"
 
-#include <avahi-client/publish.h>
 #include <avahi-common/alternative.h>
 #include <avahi-common/error.h>
 
@@ -31,7 +30,7 @@
  * SECTION:dispatcher
  * @short_description: publish DNS-SD services
  * @include: libepc/dispatcher.h
- * @stability: Private
+ * @stability: Unstable
  *
  * The #EpcDispatcher object provides an easy method for publishing
  * DNS-SD services. Unlike established APIs like Avahi or HOWL the
@@ -56,6 +55,8 @@
 
 typedef struct _EpcService EpcService;
 
+typedef void (*EpcServiceCallback) (EpcService *service);
+
 enum
 {
   PROP_NONE,
@@ -78,8 +79,6 @@ struct _EpcService
   AvahiStringList *details;
 };
 
-extern gboolean _epc_debug;
-
 /**
  * EpcDispatcherPrivate:
  * @name: the service name
@@ -90,13 +89,15 @@ extern gboolean _epc_debug;
  */
 struct _EpcDispatcherPrivate
 {
-  gchar        *name;
-  AvahiClient  *client;
-  GHashTable   *services;
+  gchar      *name;
+  GHashTable *services;
+  guint       watch_id;
 };
 
 static void epc_dispatcher_handle_collision (EpcDispatcher *self);
-static void epc_dispatcher_reset_client     (EpcDispatcher *self);
+static void epc_service_run                 (EpcService    *self);
+
+G_DEFINE_TYPE (EpcDispatcher, epc_dispatcher, G_TYPE_OBJECT);
 
 static gboolean
 epc_service_commit_cb (gpointer data)
@@ -122,12 +123,12 @@ epc_service_publish_subtype (EpcService  *self,
 {
   gint result;
 
-  if (G_UNLIKELY (_epc_debug))
+  if (EPC_DEBUG_LEVEL (1))
     g_debug ("%s: Publishing sub-service `%s' for `%s'...",
              G_STRLOC, subtype, self->dispatcher->priv->name);
 
   result = avahi_entry_group_add_service_subtype (self->group,
-                                                  AVAHI_IF_UNSPEC, 
+                                                  AVAHI_IF_UNSPEC,
                                                   self->protocol, 0,
                                                   self->dispatcher->priv->name,
                                                   self->type, self->domain,
@@ -146,7 +147,7 @@ epc_service_publish_details (EpcService *self)
 {
   gint result;
 
-  if (G_UNLIKELY (_epc_debug))
+  if (EPC_DEBUG_LEVEL (1))
     g_debug ("%s: Publishing details for `%s'...",
              G_STRLOC, self->dispatcher->priv->name);
 
@@ -167,43 +168,51 @@ epc_service_publish_details (EpcService *self)
 static void
 epc_service_publish (EpcService *self)
 {
-  gint result;
-  GList *iter;
-
-  if (G_UNLIKELY (_epc_debug))
-    g_debug ("%s: Publishing service `%s' for `%s'...",
-             G_STRLOC, self->type, self->dispatcher->priv->name);
-
-  result = avahi_entry_group_add_service_strlst (self->group,
-                                                 AVAHI_IF_UNSPEC, self->protocol, 0,
-                                                 self->dispatcher->priv->name,
-                                                 self->type, self->domain,
-                                                 self->host, self->port,
-                                                 self->details);
-
-  if (AVAHI_ERR_COLLISION == result)
-    epc_dispatcher_handle_collision (self->dispatcher);
-  else if (AVAHI_OK != result)
-    g_warning ("%s: Failed to publish service `%s' for `%s': %s (%d)",
-               G_STRLOC, self->type, self->dispatcher->priv->name,
-               avahi_strerror (result), result);
-  else
+  if (self->group)
     {
-      for (iter = self->subtypes; iter; iter = iter->next)
-        epc_service_publish_subtype (self, iter->data);
+      gint result;
+      GList *iter;
 
-      epc_service_schedule_commit (self);
+      if (EPC_DEBUG_LEVEL (1))
+        g_debug ("%s: Publishing service `%s' for `%s'...",
+                 G_STRLOC, self->type, self->dispatcher->priv->name);
+
+      result = avahi_entry_group_add_service_strlst (self->group,
+                                                     AVAHI_IF_UNSPEC, self->protocol, 0,
+                                                     self->dispatcher->priv->name,
+                                                     self->type, self->domain,
+                                                     self->host, self->port,
+                                                     self->details);
+
+      if (AVAHI_ERR_COLLISION == result)
+        epc_dispatcher_handle_collision (self->dispatcher);
+      else if (AVAHI_OK != result)
+        g_warning ("%s: Failed to publish service `%s' for `%s': %s (%d)",
+                   G_STRLOC, self->type, self->dispatcher->priv->name,
+                   avahi_strerror (result), result);
+      else
+        {
+          for (iter = self->subtypes; iter; iter = iter->next)
+            epc_service_publish_subtype (self, iter->data);
+
+          epc_service_schedule_commit (self);
+        }
     }
+  else
+    epc_service_run (self);
 }
 
 static void
 epc_service_reset (EpcService *self)
 {
-  if (G_UNLIKELY (_epc_debug))
-    g_debug ("%s: Resetting `%s' for `%s'...",
-             G_STRLOC, self->type, self->dispatcher->priv->name);
+  if (self->group)
+    {
+      if (EPC_DEBUG_LEVEL (1))
+        g_debug ("%s: Resetting `%s' for `%s'...",
+                 G_STRLOC, self->type, self->dispatcher->priv->name);
 
-  avahi_entry_group_reset (self->group);
+      avahi_entry_group_reset (self->group);
+    }
 }
 
 static void
@@ -211,8 +220,8 @@ epc_service_group_cb (AvahiEntryGroup      *group,
                       AvahiEntryGroupState  state,
                       gpointer              data)
 {
-  AvahiClient *client = avahi_entry_group_get_client (group);
   EpcService *self = data;
+  GError *error = NULL;
 
   g_assert (NULL == self->group || group == self->group);
 
@@ -233,11 +242,31 @@ epc_service_group_cb (AvahiEntryGroup      *group,
         break;
 
       case AVAHI_ENTRY_GROUP_FAILURE:
-        g_warning ("%s: Avahi group failure: %s (%d)\n", G_STRLOC,
-                   avahi_strerror (avahi_client_errno (client)),
-                   avahi_client_errno (client));
-        epc_dispatcher_reset_client (self->dispatcher);
-        break;
+        {
+          AvahiClient *client = avahi_entry_group_get_client (group);
+          gint error_code = avahi_client_errno (client);
+
+          g_warning ("%s: Failed to publish service records: %s.",
+                     G_STRFUNC, avahi_strerror (error_code));
+
+          epc_shell_restart_avahi_client (G_STRLOC);
+          break;
+        }
+    }
+
+  g_clear_error (&error);
+}
+
+static void
+epc_service_run (EpcService *self)
+{
+  if (NULL == self->group)
+    {
+      if (EPC_DEBUG_LEVEL (1))
+        g_debug ("%s: Creating service `%s' group for `%s'...",
+                 G_STRLOC, self->type, self->dispatcher->priv->name);
+
+      self->group = epc_shell_create_avahi_entry_group (epc_service_group_cb, self);
     }
 }
 
@@ -273,20 +302,25 @@ epc_service_new (EpcDispatcher *dispatcher,
   if (service > type)
     epc_service_add_subtype (self, type);
 
-  g_hash_table_insert (dispatcher->priv->services, self->type, self);
-
-  self->group = avahi_entry_group_new (dispatcher->priv->client,
-                                       epc_service_group_cb, self);
-
   return self;
 }
 
+static void
+epc_service_suspend (EpcService *self)
+{
+  if (self->group)
+    {
+      avahi_entry_group_free (self->group);
+      self->group = NULL;
+    }
+}
+
 static void
 epc_service_free (gpointer data)
 {
   EpcService *self = data;
 
-  avahi_entry_group_free (self->group);
+  epc_service_suspend (self);
   avahi_string_list_free (self->details);
 
   g_list_foreach (self->subtypes, (GFunc)g_free, NULL);
@@ -302,70 +336,67 @@ epc_service_free (gpointer data)
   g_slice_free (EpcService, self);
 }
 
-G_DEFINE_TYPE (EpcDispatcher, epc_dispatcher, G_TYPE_OBJECT);
-
 static void
-epc_dispatcher_publish_cb (gpointer key G_GNUC_UNUSED,
-                           gpointer value,
-                           gpointer data G_GNUC_UNUSED)
+epc_dispatcher_services_cb (gpointer key G_GNUC_UNUSED,
+                            gpointer value,
+                            gpointer data)
 {
-  epc_service_publish (value);
+  ((EpcServiceCallback) data) (value);
 }
 
 static void
-epc_dispatcher_reset_cb (gpointer key G_GNUC_UNUSED,
-                         gpointer value,
-                         gpointer data G_GNUC_UNUSED)
+epc_dispatcher_foreach_service (EpcDispatcher      *self,
+                                EpcServiceCallback  callback)
 {
-  epc_service_reset (value);
+  g_hash_table_foreach (self->priv->services, epc_dispatcher_services_cb, callback);
 }
 
 static void
-epc_dispatcher_client_cb (AvahiClient      *client,
+epc_dispatcher_client_cb (AvahiClient      *client G_GNUC_UNUSED,
                           AvahiClientState  state,
                           gpointer          data)
 {
   EpcDispatcher *self = data;
-
-  g_assert (NULL == self->priv->client ||
-            client == self->priv->client);
+  GError *error = NULL;
 
   switch (state)
     {
       case AVAHI_CLIENT_S_RUNNING:
-        if (G_UNLIKELY (_epc_debug))
+        if (EPC_DEBUG_LEVEL (1))
           g_debug ("%s: Avahi client is running...", G_STRLOC);
 
-        g_hash_table_foreach (self->priv->services, epc_dispatcher_publish_cb, NULL);
+        epc_dispatcher_foreach_service (self, epc_service_publish);
         break;
 
       case AVAHI_CLIENT_S_REGISTERING:
-        if (G_UNLIKELY (_epc_debug))
+        if (EPC_DEBUG_LEVEL (1))
           g_debug ("%s: Avahi client is registering...", G_STRLOC);
 
-        g_hash_table_foreach (self->priv->services, epc_dispatcher_reset_cb, self);
+        epc_dispatcher_foreach_service (self, epc_service_reset);
         break;
 
       case AVAHI_CLIENT_S_COLLISION:
-        if (G_UNLIKELY (_epc_debug))
+        if (EPC_DEBUG_LEVEL (1))
           g_debug ("%s: Collision detected...", G_STRLOC);
 
         epc_dispatcher_handle_collision (self);
         break;
 
       case AVAHI_CLIENT_FAILURE:
-        g_warning ("%s: Avahi client failure: %s (%d)\n", G_STRLOC,
-                   avahi_strerror (avahi_client_errno (client)),
-                   avahi_client_errno (client));
-        epc_dispatcher_reset_client (self);
+        if (EPC_DEBUG_LEVEL (1))
+          g_debug ("%s: Suspending entry groups...", G_STRLOC);
+
+        epc_dispatcher_foreach_service (self, epc_service_suspend);
         break;
 
       case AVAHI_CLIENT_CONNECTING:
-        if (G_UNLIKELY (_epc_debug))
+        if (EPC_DEBUG_LEVEL (1))
           g_debug ("%s: Waiting for Avahi server...", G_STRLOC);
 
         break;
     }
+
+  g_clear_error (&error);
 }
 
 static void
@@ -375,37 +406,16 @@ epc_dispatcher_handle_collision (EpcDispatcher *self)
 
   alternative = avahi_alternative_service_name (self->priv->name);
 
-  g_warning ("%s: Service name collision for `%s', renaming to `%s'.",
+  g_message ("%s: Service name collision for `%s', renaming to `%s'.",
              G_STRLOC, self->priv->name, alternative);
 
-  g_hash_table_foreach (self->priv->services, epc_dispatcher_reset_cb, self);
+  epc_dispatcher_foreach_service (self, epc_service_reset);
 
   g_free (self->priv->name);
   self->priv->name = alternative;
-
   g_object_notify (G_OBJECT (self), "name");
-  g_hash_table_foreach (self->priv->services, epc_dispatcher_publish_cb, self);
-}
 
-static void
-epc_dispatcher_reset_client (EpcDispatcher *self)
-{
-  GError *error = NULL;
-
-  if (self->priv->client)
-    {
-      avahi_client_free (self->priv->client);
-      self->priv->client = NULL;
-    }
-
-  self->priv->client = epc_shell_create_avahi_client (AVAHI_CLIENT_NO_FAIL,
-                                                      epc_dispatcher_client_cb, 
-                                                      self, &error);
-
-  if (NULL == self->priv->client)
-    g_warning ("%s: %s", G_STRFUNC, error->message);
-
-  g_clear_error (&error);
+  epc_dispatcher_foreach_service (self, epc_service_publish);
 }
 
 static void
@@ -417,9 +427,6 @@ epc_dispatcher_init (EpcDispatcher *self)
 
   self->priv->services = g_hash_table_new_full (g_str_hash, g_str_equal,
                                                 NULL, epc_service_free);
-
-  epc_shell_ref ();
-  epc_dispatcher_reset_client (self);
 }
 
 static void
@@ -434,14 +441,14 @@ epc_dispatcher_set_property (GObject      *object,
     {
       case PROP_NAME:
         g_return_if_fail (NULL != g_value_get_string (value));
-  
+
         g_free (self->priv->name);
         self->priv->name = g_value_dup_string (value);
 
-        /* The reset also causes a transition into the UNCOMMITED state, 
+        /* The reset also causes a transition into the UNCOMMITED state,
          * which causes re-publication of the services.
          */
-        g_hash_table_foreach (self->priv->services, epc_dispatcher_reset_cb, self);
+        epc_dispatcher_foreach_service (self, epc_service_reset);
         break;
 
       default:
@@ -481,10 +488,10 @@ epc_dispatcher_dispose (GObject *object)
       self->priv->services = NULL;
     }
 
-  if (self->priv->client)
+  if (self->priv->watch_id)
     {
-      avahi_client_free (self->priv->client);
-      self->priv->client = NULL;
+      epc_shell_watch_remove (self->priv->watch_id);
+      self->priv->watch_id = 0;
     }
 
   g_free (self->priv->name);
@@ -493,13 +500,6 @@ epc_dispatcher_dispose (GObject *object)
   G_OBJECT_CLASS (epc_dispatcher_parent_class)->dispose (object);
 }
 
-static void
-epc_dispatcher_finalize (GObject *object)
-{
-  G_OBJECT_CLASS (epc_dispatcher_parent_class)->finalize (object);
-  epc_shell_unref ();
-}
-
 static void
 epc_dispatcher_class_init (EpcDispatcherClass *cls)
 {
@@ -507,7 +507,6 @@ epc_dispatcher_class_init (EpcDispatcherClass *cls)
 
   oclass->set_property = epc_dispatcher_set_property;
   oclass->get_property = epc_dispatcher_get_property;
-  oclass->finalize = epc_dispatcher_finalize;
   oclass->dispose = epc_dispatcher_dispose;
 
   g_object_class_install_property (oclass, PROP_NAME,
@@ -537,6 +536,33 @@ epc_dispatcher_new (const gchar *name)
   return g_object_new (EPC_TYPE_DISPATCHER, "name", name, NULL);
 }
 
+/**
+ * epc_dispatcher_run:
+ * @dispatcher: a #EpcDispatcher
+ * @error: return location for a #GError, or %NULL
+ *
+ * Starts the <citetitle>Avahi</citetitle> client of the #EpcDispatcher. If the
+ * client was not started, the function returns %FALSE and sets @error. The
+ * error domain is #EPC_AVAHI_ERROR. Possible error codes are those of the
+ * <citetitle>Avahi</citetitle> library.
+ *
+ * Returns: %TRUE when the dispatcher was started successfully,
+ * %FALSE if an error occurred.
+ */
+gboolean
+epc_dispatcher_run (EpcDispatcher  *self,
+                    GError        **error)
+{
+  g_return_val_if_fail (EPC_IS_DISPATCHER (self), FALSE);
+  g_return_val_if_fail (0 == self->priv->watch_id, FALSE);
+
+  self->priv->watch_id =
+    epc_shell_watch_avahi_client_state (epc_dispatcher_client_cb,
+                                        self, NULL, error);
+
+  return (0 != self->priv->watch_id);
+}
+
 /**
  * epc_dispatcher_reset:
  * @dispatcher: a #EpcDispatcher
@@ -571,7 +597,7 @@ epc_dispatcher_reset (EpcDispatcher *self)
  * The function can be called more than once. Is this necessary when the server
  * provides different access methods. For instance a web server could provide
  * HTTP and encrypted HTTPS services at the same time. Calling this function
- * multiple times also is useful for servers providing the same service at 
+ * multiple times also is useful for servers providing the same service at
  * different, but not all network interfaces of the host.
  *
  * When passing %NULL for @domain, the service is announced within the local
@@ -579,7 +605,7 @@ epc_dispatcher_reset (EpcDispatcher *self)
  * responsible server must be <ulink url="http://www.dns-sd.org/ServerSetup.html">
  * configured to support DNS-SD</ulink>.
  *
- * Pass %NULL for @host to use the official host name of the machine to announce 
+ * Pass %NULL for @host to use the official host name of the machine to announce
  * the service. On machines with multiple DNS entries you might want to explictly
  * choose a fully qualified DNS name to announce the service.
  */
@@ -608,6 +634,11 @@ epc_dispatcher_add_service (EpcDispatcher    *self,
                              type, domain, host, port, args);
 
   va_end (args);
+
+  g_hash_table_insert (self->priv->services, service->type, service);
+
+  if (self->priv->watch_id)
+    epc_service_run (service);
 }
 
 /**
@@ -639,7 +670,9 @@ epc_dispatcher_add_service_subtype (EpcDispatcher *self,
   g_return_if_fail (NULL != service);
 
   epc_service_add_subtype (service, subtype);
-  epc_service_publish_subtype (service, subtype);
+
+  if (self->priv->watch_id)
+    epc_service_publish_subtype (service, subtype);
 }
 
 /**
@@ -715,20 +748,3 @@ epc_dispatcher_get_name (EpcDispatcher *self)
   g_return_val_if_fail (EPC_IS_DISPATCHER (self), NULL);
   return self->priv->name;
 }
-
-/**
- * epc_dispatcher_get_host_name:
- * @dispatcher: a #EpcDispatcher
- *
- * Query the official host name of this machine.
- *
- * Returns: The official host name, or %NULL on error.
- */
-G_CONST_RETURN gchar*
-epc_dispatcher_get_host_name (EpcDispatcher *self)
-{
-  g_return_val_if_fail (EPC_IS_DISPATCHER (self), NULL);
-  g_return_val_if_fail (NULL != self->priv->client, NULL);
-  return avahi_client_get_host_name (self->priv->client);
-}
-
