@@ -1,7 +1,11 @@
 #!/bin/bash
 
 test -n "$EPC_DEBUG" && echo "$0: running $1..."
+
+EPC_DEBUG="${EPC_DEBUG:-1}" \
+G_DEBUG="${G_DEBUG:-fatal-warnings}" \
 "$@" 2> "$1.err" > "$1.out"
+
 result=$?
 
 if test -n "$EPC_DEBUG"
@@ -13,9 +17,6 @@ then
   echo "===== BEGIN OF STDERR ====="
   cat "$1.out"
   echo "===== END OF STDERR ====="
-elif test $result -ne 0
-then
-  cat "$1.err"
 fi
 
 exit $result
